@@ -13,6 +13,9 @@
 #include "Ppmd7.h"
 #include "Ppmd8.h"
 
+#include "Buffer.h"
+#include "Ppmd8Tdecoder.h"
+
 #ifndef Py_UNREACHABLE
     #define Py_UNREACHABLE() assert(0)
 #endif
@@ -20,6 +23,7 @@
 /* ----------------------------
      BlocksOutputBuffer code
    ---------------------------- */
+
 typedef struct {
     /* List of blocks */
     PyObject *list;
@@ -29,19 +33,6 @@ typedef struct {
     Py_ssize_t max_length;
 } BlocksOutputBuffer;
 
-typedef struct {
-    /* Inherits from IByteOut */
-    void (*Write)(void *p, Byte b);
-    BlocksOutputBuffer * buffer;
-    PPMD_outBuffer *outBuffer;
-} BufferWriter;
-
-typedef struct {
-    /* Inherits from IByteIn */
-    Byte (*Read)(void *p);
-    PPMD_inBuffer *inBuffer;
-} BufferReader;
-
 static const char unable_allocate_msg[] = "Unable to allocate output buffer.";
 
 /* Block size sequence. Below functions assume the type is int. */
@@ -49,9 +40,9 @@ static const char unable_allocate_msg[] = "Unable to allocate output buffer.";
 #define KB (1024)
 #define MB (1024*1024)
 static const int BUFFER_BLOCK_SIZE[] =
-        { 32*KB, 64*KB, 256*KB, 1*MB, 4*MB, 8*MB, 16*MB, 16*MB,
-          32*MB, 32*MB, 32*MB, 32*MB, 64*MB, 64*MB, 128*MB, 128*MB,
-          256*MB };
+        {32 * KB, 64 * KB, 256 * KB, 1 * MB, 4 * MB, 8 * MB, 16 * MB, 16 * MB,
+         32 * MB, 32 * MB, 32 * MB, 32 * MB, 64 * MB, 64 * MB, 128 * MB, 128 * MB,
+         256 * MB};
 
 /* According to the block sizes defined by BUFFER_BLOCK_SIZE, the whole
    allocated size growth step is
@@ -87,9 +78,8 @@ static const int BUFFER_BLOCK_SIZE[] =
    Return -1 on failure
 */
 static inline int
-OutputBuffer_InitAndGrow(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob,
-                         Py_ssize_t max_length)
-{
+OutputBuffer_InitAndGrow(BlocksOutputBuffer *buffer, OutBuffer *ob,
+                         Py_ssize_t max_length) {
     PyObject *b;
     int block_size;
 
@@ -131,9 +121,8 @@ OutputBuffer_InitAndGrow(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob,
    Return -1 on failure
 */
 static inline int
-OutputBuffer_InitWithSize(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob,
-                          Py_ssize_t init_size)
-{
+OutputBuffer_InitWithSize(BlocksOutputBuffer *buffer, OutBuffer *ob,
+                          Py_ssize_t init_size) {
     PyObject *b;
 
     /* The first block */
@@ -167,8 +156,7 @@ OutputBuffer_InitWithSize(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob,
    Return -1 on failure
 */
 static inline int
-OutputBuffer_Grow(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob)
-{
+OutputBuffer_Grow(BlocksOutputBuffer *buffer, OutBuffer *ob) {
     PyObject *b;
     const Py_ssize_t list_len = Py_SIZE(buffer->list);
     int block_size;
@@ -225,28 +213,26 @@ OutputBuffer_Grow(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob)
 /* Whether the output data has reached max_length.
    The avail_out must be 0, please check it before calling. */
 static inline int
-OutputBuffer_ReachedMaxLength(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob)
-{
+OutputBuffer_ReachedMaxLength(BlocksOutputBuffer *buffer, OutBuffer *ob) {
     /* Ensure (data size == allocated size) */
     assert(ob->pos == ob->size);
 
     return buffer->allocated == buffer->max_length;
 }
 
+
 /* Finish the buffer.
    Return a bytes object on success
    Return NULL on failure
 */
 static PyObject *
-OutputBuffer_Finish(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob)
-{
+OutputBuffer_Finish(BlocksOutputBuffer *buffer, OutBuffer *ob) {
     PyObject *result, *block;
     const Py_ssize_t list_len = Py_SIZE(buffer->list);
 
     /* Fast path for single block */
     if ((list_len == 1 && ob->pos == ob->size) ||
-        (list_len == 2 && ob->pos == 0))
-    {
+    (list_len == 2 && ob->pos == 0)) {
         block = PyList_GET_ITEM(buffer->list, 0);
         Py_INCREF(block);
 
@@ -267,7 +253,7 @@ OutputBuffer_Finish(BlocksOutputBuffer *buffer, PPMD_outBuffer *ob)
 
         /* Blocks except the last one */
         Py_ssize_t i = 0;
-        for (; i < list_len-1; i++) {
+        for (; i < list_len - 1; i++) {
             block = PyList_GET_ITEM(buffer->list, i);
             memcpy(posi, PyBytes_AS_STRING(block), Py_SIZE(block));
             posi += Py_SIZE(block);
@@ -295,36 +281,9 @@ OutputBuffer_OnError(BlocksOutputBuffer *buffer)
    End of BlocksOutputBuffer code
    ------------------------------ */
 
-static void Write(void *p, Byte b) {
-    BufferWriter *bufferWriter = (BufferWriter *) p;
-
-    // Grow output buffer size when buffer is full
-    if (bufferWriter->outBuffer->size == bufferWriter->outBuffer->pos) {
-        if (OutputBuffer_Grow(bufferWriter->buffer, bufferWriter->outBuffer) < 0) {
-            // FIXME: propagate memory error to upstream
-            goto error;
-        }
-    }
-
-    *((Byte *)bufferWriter->outBuffer->dst + bufferWriter->outBuffer->pos++) = b;
-    return;
-error:
-    return;
-}
-
-Byte Reader(void *p) {
-    BufferReader *bufferReader = (BufferReader *) p;
-    PPMD_inBuffer *inBuffer = bufferReader->inBuffer;
-    if (inBuffer->pos == inBuffer->size) {
-        return -1;
-    } else {
-        return *((const Byte *)inBuffer->src + inBuffer->pos++);
-    }
-}
-
 static ISzAlloc allocator = {
-    PyMem_Malloc,
-    PyMem_Free
+        PyMem_Malloc,
+        PyMem_Free
 };
 
 typedef struct {
@@ -414,10 +373,15 @@ typedef struct {
     /* 1 when end mark observed */
     char eof;
 
+    /* Output Buffer */
+    BlocksOutputBuffer *blocksOutputBuffer;
+
     /* __init__ has been called, 0 or 1. */
     char inited;
     /* decode has been called with some data*/
     char inited2;
+    /* threaded decoder context */
+    ppmd8_args *args;
 } Ppmd8Decoder;
 
 typedef struct {
@@ -442,20 +406,21 @@ static const char init_twice_msg[] = "__init__ method is called twice.";
 static const char flush_twice_msg[] = "flush method is called twice.";
 
 static inline void
-clamp_max_order(unsigned long *max_order) {
-    if (*max_order < 2) {
-        *max_order = 2;
-    } else if (*max_order > 64) {
-        *max_order = 64;
+clamp_max_order(unsigned long *max_order, unsigned long max) {
+    assert(PPMD7_MIN_ORDER == PPMD8_MIN_ORDER);
+    if (*max_order < PPMD7_MIN_ORDER) {
+        *max_order = PPMD7_MIN_ORDER;
+    } else if (*max_order > max) {
+        *max_order = max;
     }
 }
 
 static inline void
 clamp_memory_size(unsigned long *memorySize) {
-    if (*memorySize < 1 << 11) {
-        *memorySize = 1 << 11;
-    } else if (*memorySize > 0xFFFFFFFF - 12 * 3) {
-        *memorySize = 0xFFFFFFFF - 12 * 3;
+    if (*memorySize < PPMD7_MIN_MEM_SIZE) {
+        *memorySize = PPMD7_MIN_MEM_SIZE;
+    } else if (*memorySize > PPMD7_MAX_MEM_SIZE) {
+        *memorySize = PPMD7_MAX_MEM_SIZE;
     }
 }
 
@@ -541,7 +506,7 @@ Ppmd7Decoder_init(Ppmd7Decoder *self, PyObject *args, PyObject *kwargs)
                 goto error;
             }
         }
-        clamp_max_order(&maximum_order);
+        clamp_max_order(&maximum_order, PPMD7_MAX_ORDER);
     }
 
     if (mem_size != Py_None) {
@@ -583,8 +548,8 @@ Ppmd7Decoder_flush(Ppmd7Decoder *self, PyObject *args, PyObject *kwargs) {
     static char *kwlist[] = {"length", NULL};
     BlocksOutputBuffer buffer;
     int length;
-    PPMD_inBuffer in;
-    PPMD_outBuffer out;
+    InBuffer in;
+    OutBuffer out;
     PyObject *ret = NULL;
     BufferReader reader;
 
@@ -630,7 +595,7 @@ Ppmd7Decoder_flush(Ppmd7Decoder *self, PyObject *args, PyObject *kwargs) {
         in.pos = 0;
     }
 
-    reader.Read = Reader;
+    reader.Read = (Byte (*)(void *)) Reader;
     reader.inBuffer = &in;
     self->rangeDec->Stream = (IByteIn *) &reader;
 
@@ -677,8 +642,8 @@ Ppmd7Decoder_decode(Ppmd7Decoder *self,  PyObject *args, PyObject *kwargs) {
     BlocksOutputBuffer buffer;
     Py_buffer data;
     int length;
-    PPMD_inBuffer in;
-    PPMD_outBuffer out;
+    InBuffer in;
+    OutBuffer out;
     PyObject *ret = NULL;
     char use_input_buffer;
     BufferReader reader;
@@ -777,7 +742,7 @@ Ppmd7Decoder_decode(Ppmd7Decoder *self,  PyObject *args, PyObject *kwargs) {
     }
     assert(in.pos == 0);
 
-    reader.Read = Reader;
+    reader.Read = (Byte (*)(void *)) Reader;
     reader.inBuffer = &in;
     self->rangeDec->Stream = (IByteIn *) &reader;
 
@@ -806,7 +771,6 @@ Ppmd7Decoder_decode(Ppmd7Decoder *self,  PyObject *args, PyObject *kwargs) {
     assert(self->inited2 == 1);
 
     Bool result = True;
-    Py_BEGIN_ALLOW_THREADS
     if (data.len  > 0) {
         for (int i = 0; i < length; i++) {
             if (in.pos == in.size) {
@@ -819,10 +783,11 @@ Ppmd7Decoder_decode(Ppmd7Decoder *self,  PyObject *args, PyObject *kwargs) {
                     break;
                 }
             }
+            Py_BEGIN_ALLOW_THREADS
             *((Byte *)out.dst + out.pos++) = Ppmd7_DecodeSymbol(self->cPpmd7, self->rangeDec);
+            Py_END_ALLOW_THREADS
         }
     }
-    Py_END_ALLOW_THREADS
 
     if (!result)
         goto error;
@@ -999,7 +964,7 @@ Ppmd7Encoder_init(Ppmd7Encoder *self, PyObject *args, PyObject *kwargs)
                 goto error;
             }
         }
-        clamp_max_order(&maximum_order);
+        clamp_max_order(&maximum_order, PPMD7_MAX_ORDER);
     }
 
     if (mem_size != Py_None) {
@@ -1044,7 +1009,7 @@ Ppmd7Encoder_encode(Ppmd7Encoder *self,  PyObject *args, PyObject *kwargs) {
     BlocksOutputBuffer buffer;
     Py_buffer data;
     PyObject *ret;
-    PPMD_outBuffer out;
+    OutBuffer out;
     BufferWriter writer;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
@@ -1059,14 +1024,15 @@ Ppmd7Encoder_encode(Ppmd7Encoder *self,  PyObject *args, PyObject *kwargs) {
         goto error;
     }
 
-    writer.Write = Write;
+    writer.Write = (void (*)(void *, Byte)) Writer;
     writer.outBuffer = &out;
     self->rangeEnc->Stream = (IByteOut *) &writer;
 
     Bool result = True;
-    Py_BEGIN_ALLOW_THREADS
     for (UInt32 i = 0; i < data.len; i++){
+        Py_BEGIN_ALLOW_THREADS
         Ppmd7_EncodeSymbol(self->cPpmd7, self->rangeEnc, *((Byte *)data.buf + i));
+        Py_END_ALLOW_THREADS
         if (out.size == out.pos) {
             if (OutputBuffer_Grow(&buffer, &out) < 0) {
                 PyErr_SetString(PyExc_ValueError, "No memory.");
@@ -1077,7 +1043,6 @@ Ppmd7Encoder_encode(Ppmd7Encoder *self,  PyObject *args, PyObject *kwargs) {
             }
         }
     }
-    Py_END_ALLOW_THREADS
     if (!result)
         goto error;
 
@@ -1100,7 +1065,7 @@ Ppmd7Encoder_flush(Ppmd7Encoder *self, PyObject *args, PyObject *kwargs)
 {
     PyObject *ret;
     CPpmd7z_RangeEnc *rc = self->rangeEnc;
-    PPMD_outBuffer out;
+    OutBuffer out;
     BlocksOutputBuffer buffer;
     BufferWriter writer;
 
@@ -1115,7 +1080,7 @@ Ppmd7Encoder_flush(Ppmd7Encoder *self, PyObject *args, PyObject *kwargs)
         goto error;
     }
 
-    writer.Write = Write;
+    writer.Write = (void (*)(void *, Byte)) Writer;
     writer.outBuffer = &out;
     rc->Stream = (IByteOut *) &writer;
 
@@ -1190,14 +1155,21 @@ Ppmd8Decoder_dealloc(Ppmd8Decoder *self)
     if (self->lock) {
         PyThread_free_lock(self->lock);
     }
-    Ppmd8_Free(self->cPpmd8, &allocator);
+    Ppmd8T_Free(self->cPpmd8, self->args, &allocator);
+    BufferReader *bufferReader = (BufferReader *)self->cPpmd8->Stream.In;
+    PyMem_Free(bufferReader->inBuffer);
+    PyMem_Free(bufferReader);
+    PyMem_Free(self->args->out);
+    PyMem_Free(self->blocksOutputBuffer);
+    PyMem_Free(self->args);
+    PyMem_Free(self->cPpmd8);
     PyTypeObject *tp = Py_TYPE(self);
     tp->tp_free((PyObject*)self);
     Py_DECREF(tp);
 }
 
 PyDoc_STRVAR(Ppmd8Decoder_doc, "A PPMd compression algorithm decoder.\n\n"
-                                 "Ppmd8Decoder.__init__(self, max_order, mem_size)\n"
+                                 "Ppmd8Decoder.__init__(self, max_order, mem_size, restore_method=0)\n"
                                  "----\n"
                                  "Initialize a Ppmd8Decoder object.\n\n"
                                  "Arguments\n"
@@ -1207,18 +1179,25 @@ PyDoc_STRVAR(Ppmd8Decoder_doc, "A PPMd compression algorithm decoder.\n\n"
                                  "mem_size:  max memory size in bytes the compressor is able to use, bigger values improve compression,\n"
                                  "           raging from 10kB to physical memory size.\n"
                                  "           Default size is 16MB.\n"
+                                 "restore_method: restore method, 0=restart, 1=cutoff.\n"
                                  );
 
 static int
 Ppmd8Decoder_init(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs)
 {
-    static char *kwlist[] = {"max_order", "mem_size", NULL};
+    static char *kwlist[] = {"max_order", "mem_size", "restore_method", "endmark", NULL};
     PyObject *max_order = Py_None;
     PyObject *mem_size = Py_None;
+    int restore_method = PPMD8_RESTORE_METHOD_RESTART;
+    int endmark = True;
+    BlocksOutputBuffer *blocksOutputBuffer;
+    BufferReader *bufferReader;
+    InBuffer *in;
+    OutBuffer *out;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
-                                     "OO:Ppmd8Decoder.__init__", kwlist,
-                                     &max_order, &mem_size)) {
+                                     "OO|ii:Ppmd8Decoder.__init__", kwlist,
+                                     &max_order, &mem_size, &restore_method, &endmark)) {
         return -1;
     }
 
@@ -1242,7 +1221,7 @@ Ppmd8Decoder_init(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs)
                 goto error;
             }
         }
-        clamp_max_order(&maximum_order);
+        clamp_max_order(&maximum_order, PPMD8_MAX_ORDER);
     }
 
     if (mem_size != Py_None) {
@@ -1257,13 +1236,61 @@ Ppmd8Decoder_init(Ppmd8Decoder *self, PyObject *args, PyObject *kwargs)
         clamp_memory_size(&memory_size);
     }
 
+    bufferReader = PyMem_Malloc(sizeof(BufferReader));
+    if (bufferReader == NULL) {
+        PyErr_NoMemory();
+        goto error;
+    }
+    blocksOutputBuffer = PyMem_Malloc(sizeof(BlocksOutputBuffer));
+    if (blocksOutputBuffer == NULL) {
+        PyMem_Free(bufferReader);
+        PyErr_NoMemory();
+        goto error;
+    }
+    in = PyMem_Malloc(sizeof(InBuffer));
+    if (in == NULL) {
+        PyMem_Free(bufferReader);
+        PyMem_Free(blocksOutputBuffer);
+        PyErr_NoMemory();
+        goto error;
+    }
+    out = PyMem_Malloc(sizeof(OutBuffer));
+    if (out == NULL) {
+        PyMem_Free(in);
+        PyMem_Free(blocksOutputBuffer);
+        PyMem_Free(bufferReader);
+        PyErr_NoMemory();
+        goto error;
+    }
+    self->args = PyMem_Malloc(sizeof(ppmd8_args));
+    if (self->args == NULL) {
+        PyMem_Free(out);
+        PyMem_Free(in);
+        PyMem_Free(blocksOutputBuffer);
+        PyMem_Free(bufferReader);
+        PyErr_NoMemory();
+        goto error;
+    }
+
     if ((self->cPpmd8 = PyMem_Malloc(sizeof(CPpmd8))) != NULL) {
         Ppmd8_Construct(self->cPpmd8);
-        if (Ppmd8_Alloc(self->cPpmd8, (UInt32)memory_size, &allocator)) {
-            Ppmd8_Init(self->cPpmd8, (unsigned int)maximum_order, PPMD8_RESTORE_METHOD_RESTART);
+        if (Ppmd8_Alloc(self->cPpmd8, memory_size ,&allocator)) {
+            Ppmd8_Init(self->cPpmd8, maximum_order, restore_method);
+            bufferReader->Read = (Byte (*)(void *)) TReader;
+            bufferReader->inBuffer = in;
+            self->cPpmd8->Stream.In = (IByteIn *) bufferReader;
+            self->args->cPpmd8 = self->cPpmd8;
+            self->args->endmark = endmark;
+            self->blocksOutputBuffer = blocksOutputBuffer;
+            self->args->out = out;
             goto success;
         }
         PyMem_Free(self->cPpmd8);
+        PyMem_Free(self->args);
+        PyMem_Free(out);
+        PyMem_Free(in);
+        PyMem_Free(blocksOutputBuffer);
+        PyMem_Free(bufferReader);
         PyErr_NoMemory();
     }
 
@@ -1308,14 +1335,10 @@ PyDoc_STRVAR(Ppmd8Decoder_decode_doc, "decode()\n"
 static PyObject *
 Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
     static char *kwlist[] = {"data", "length", NULL};
-    BlocksOutputBuffer buffer;
     Py_buffer data;
     int length = -1;
-    PPMD_inBuffer in;
-    PPMD_outBuffer out;
     PyObject *ret = NULL;
     char use_input_buffer;
-    BufferReader reader;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
                                      "y*|i:Ppmd8Decoder.decode", kwlist,
@@ -1331,23 +1354,27 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
 
     ACQUIRE_LOCK(self);
 
+    BufferReader *bufferReader = (BufferReader *) self->cPpmd8->Stream.In;
+    InBuffer *in = bufferReader->inBuffer;
+    OutBuffer *out = self->args->out;
+
     /* Prepare input buffer w/wo unconsumed data */
     if (self->in_begin == self->in_end) {
         /* No unconsumed data */
         use_input_buffer = 0;
 
-        in.src = data.buf;
-        in.size = data.len;
-        in.pos = 0;
+        in->src = data.buf;
+        in->size = data.len;
+        in->pos = 0;
     } else if (data.len == 0) {
         /* Has unconsumed data, fast path for b'' */
         assert(self->in_begin < self->in_end);
 
         use_input_buffer = 1;
 
-        in.src = self->input_buffer + self->in_begin;
-        in.size = self->in_end - self->in_begin;
-        in.pos = 0;
+        in->src = self->input_buffer + self->in_begin;
+        in->size = self->in_end - self->in_begin;
+        in->pos = 0;
     } else {
         /* Has unconsumed data */
         use_input_buffer = 1;
@@ -1405,17 +1432,13 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
         /* Copy data to input buffer */
         memcpy(self->input_buffer + self->in_end, data.buf, data.len);
         self->in_end += data.len;
-        in.src = self->input_buffer + self->in_begin;
-        in.size = used_now + data.len;
-        in.pos = 0;
+        in->src = self->input_buffer + self->in_begin;
+        in->size = used_now + data.len;
+        in->pos = 0;
     }
-    assert(in.pos == 0);
+    assert(in->pos == 0);
 
-    reader.Read = Reader;
-    reader.inBuffer = &in;
-    self->cPpmd8->Stream.In = (IByteIn *) &reader;
-
-    if (OutputBuffer_InitAndGrow(&buffer, &out, length) < 0) {
+    if (OutputBuffer_InitAndGrow(self->blocksOutputBuffer, out, length) < 0) {
         PyErr_SetString(PyExc_ValueError, "No Memory.");
         RELEASE_LOCK(self);
         return NULL;
@@ -1427,53 +1450,52 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
         if (!Ppmd8_RangeDec_Init(self->cPpmd8)) {
             RELEASE_LOCK(self);
             return NULL;
-        } else {
-            self->inited2 = 1;
         }
+        self->inited2++;
+        // first time to call `decode`
+        self->args->finished = True;
     }
-    assert(self->inited2 == 1);
 
-    Bool result = True;
-    Py_BEGIN_ALLOW_THREADS
     if (data.len  > 0) {
-        int i = 0;
-        while (length == -1 || i++ < length ) {
-            if (in.pos == in.size) {
-                break;
+        int result;
+        int remains = length >= 0 ? length : INT_MAX;
+        while (True) {
+            Py_BEGIN_ALLOW_THREADS
+            result = Ppmd8T_decode(self->cPpmd8, out, remains, self->args);
+            Py_END_ALLOW_THREADS
+            if (result < 0) {
+                break; // error or eof
             }
-            if (out.pos == out.size) {
-                if (OutputBuffer_Grow(&buffer, &out) < 0) {
-                    PyErr_SetString(PyExc_ValueError, "L616: Unknown status");
-                    result = False;
+            if (result == 0) {
+                // input empty or no output data
+                if (in->pos == in->size) {
                     break;
                 }
             }
-            unsigned char c = Ppmd8_DecodeSymbol(self->cPpmd8);
-            if (c != 0x01) {
-                *((Byte *)out.dst + out.pos++) = c;
-            } else {
-                c = Ppmd8_DecodeSymbol(self->cPpmd8);
-                if (c == 0x01) { // escaped character
-                    *((Byte *)out.dst + out.pos++) = c;
-                } else if (c == 0x00){ // endmark
-                    self->eof = True;
-                    break;
-                } else {
-                    PyErr_SetString(PyExc_ValueError, "Corrupted input data.");
-                    result = False;
+            if ((remains -= result) == 0) {
+                 break;
+             }
+            if (out->pos == out->size) {
+                if (OutputBuffer_Grow(self->blocksOutputBuffer, out) < 0) {
+                    PyErr_SetString(PyExc_ValueError, "L616: Unknown status");
+                    result = -2;
                     break;
                 }
             }
         }
+        if (result == -1) {
+            self->eof = True;
+        }
+        if (result == -2) {
+            PyErr_SetString(PyExc_ValueError, "Corrupted input data.");
+            goto error;
+        }
     }
-    Py_END_ALLOW_THREADS
-    if (!result)
-        goto error;
 
-    ret = OutputBuffer_Finish(&buffer, &out);
+    ret = OutputBuffer_Finish(self->blocksOutputBuffer, out);
 
     /* Unconsumed input data */
-    if (in.pos == in.size) {
+    if (in->pos == in->size) {
         if (use_input_buffer) {
             /* Clear input_buffer */
             self->in_begin = 0;
@@ -1485,7 +1507,7 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
             self->needs_input = True;
         }
     } else {
-        const size_t data_size = in.size - in.pos;
+        const size_t data_size = in->size - in->pos;
         self->needs_input = False;
         if (!use_input_buffer) {
             /* Discard buffer if it's too small
@@ -1508,12 +1530,12 @@ Ppmd8Decoder_decode(Ppmd8Decoder *self,  PyObject *args, PyObject *kwargs) {
             }
 
             /* Copy unconsumed data */
-            memcpy(self->input_buffer, (char*)in.src + in.pos, data_size);
+            memcpy(self->input_buffer, (char*)in->src + in->pos, data_size);
             self->in_begin = 0;
             self->in_end = data_size;
         } else {
             /* Use input buffer */
-            self->in_begin += in.pos;
+            self->in_begin += in->pos;
         }
     }
     goto success;
@@ -1614,7 +1636,7 @@ Ppmd8Encoder_dealloc(Ppmd8Encoder *self)
 }
 
 PyDoc_STRVAR(Ppmd8Encoder_doc, "A PPMd compression algorithm.\n\n"
-                                 "Ppmd8Encoder.__init__(self, max_order, mem_size)\n"
+                                 "Ppmd8Encoder.__init__(self, max_order, mem_size, restore_method=0)\n"
                                  "----\n"
                                  "Initialize a Ppmd8Encoder object.\n\n"
                                  "Arguments\n"
@@ -1623,18 +1645,21 @@ PyDoc_STRVAR(Ppmd8Encoder_doc, "A PPMd compression algorithm.\n\n"
                                  "mem_size:  max memory size in bytes the compressor is able to use, bigger values improve compression,\n"
                                  "           raging from 10kB to physical memory size.\n"
                                  "           Default size is 16MB.\n"
+                                 "restore_method: restore method, 0=restart, 1=cutoff.\n"
                                  );
 
 static int
 Ppmd8Encoder_init(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
 {
-    static char *kwlist[] = {"max_order", "mem_size", NULL};
+    static char *kwlist[] = {"max_order", "mem_size", "restore_method", "endmark", NULL};
     PyObject *max_order = Py_None;
     PyObject *mem_size = Py_None;
+    int restore_method = PPMD8_RESTORE_METHOD_RESTART;
+    Bool endmark = True;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
-                                     "OO:Ppmd8Encoder.__init__", kwlist,
-                                     &max_order, &mem_size)) {
+                                     "OO|ii:Ppmd8Encoder.__init__", kwlist,
+                                     &max_order, &mem_size, &restore_method, &endmark)) {
         goto error;
     }
 
@@ -1657,7 +1682,7 @@ Ppmd8Encoder_init(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
                 goto error;
             }
         }
-        clamp_max_order(&maximum_order);
+        clamp_max_order(&maximum_order, PPMD8_MAX_ORDER);
     }
 
     if (mem_size != Py_None) {
@@ -1672,11 +1697,12 @@ Ppmd8Encoder_init(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
         clamp_memory_size(&memory_size);
     }
 
+    self->endmark = endmark;
     if ((self->cPpmd8 =  PyMem_Malloc(sizeof(CPpmd8))) != NULL) {
         Ppmd8_Construct(self->cPpmd8);
         if (Ppmd8_Alloc(self->cPpmd8, (UInt32)memory_size, &allocator)) {
             Ppmd8_RangeEnc_Init(self->cPpmd8);
-            Ppmd8_Init(self->cPpmd8, (unsigned int)maximum_order, PPMD8_RESTORE_METHOD_RESTART);
+            Ppmd8_Init(self->cPpmd8, (unsigned int)maximum_order, restore_method);
             goto success;
         }
         PyMem_Free(self->cPpmd8);
@@ -1699,7 +1725,7 @@ Ppmd8Encoder_encode(Ppmd8Encoder *self,  PyObject *args, PyObject *kwargs) {
     BlocksOutputBuffer buffer;
     Py_buffer data;
     PyObject *ret;
-    PPMD_outBuffer out;
+    OutBuffer out;
     BufferWriter writer;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwargs,
@@ -1714,27 +1740,32 @@ Ppmd8Encoder_encode(Ppmd8Encoder *self,  PyObject *args, PyObject *kwargs) {
         goto error;
     }
 
-    writer.Write = Write;
+    writer.Write = (void (*)(void *, Byte)) Writer;
     writer.outBuffer = &out;
-    self->cPpmd8->Stream.Out = (IByteOut *) &writer;
+    self->cPpmd8->Stream.Out = (IByteOut *)&writer;
 
     Bool result = True;
-    Py_BEGIN_ALLOW_THREADS
     for (UInt32 i = 0; i < data.len; i++){
-        // when endmark mode, escape 0x01.
-        if (*((Byte *)data.buf + i) == 0x01) {
-            Ppmd8_EncodeSymbol(self->cPpmd8, 0x01);
-            if (out.size == out.pos) {
-                if (OutputBuffer_Grow(&buffer, &out) < 0) {
-                    PyErr_SetString(PyExc_ValueError, "No memory.");
-                    result = False;
-                    break;
-                } else {
-                    writer.outBuffer = &out;
+        if (self->endmark) {
+            // when endmark mode, escape 0x01.
+            if (*((Byte *)data.buf + i) == 0x01) {
+                Py_BEGIN_ALLOW_THREADS
+                Ppmd8_EncodeSymbol(self->cPpmd8, 0x01);
+                Py_END_ALLOW_THREADS
+                if (out.size == out.pos) {
+                    if (OutputBuffer_Grow(&buffer, &out) < 0) {
+                        PyErr_SetString(PyExc_ValueError, "No memory.");
+                        result = False;
+                        break;
+                    } else {
+                        writer.outBuffer = &out;
+                    }
                 }
             }
         }
+        Py_BEGIN_ALLOW_THREADS
         Ppmd8_EncodeSymbol(self->cPpmd8, *((Byte *)data.buf + i));
+        Py_END_ALLOW_THREADS
         if (out.size == out.pos) {
             if (OutputBuffer_Grow(&buffer, &out) < 0) {
                 PyErr_SetString(PyExc_ValueError, "No memory.");
@@ -1745,7 +1776,6 @@ Ppmd8Encoder_encode(Ppmd8Encoder *self,  PyObject *args, PyObject *kwargs) {
             }
         }
     }
-    Py_END_ALLOW_THREADS
     if (!result)
         goto error;
 
@@ -1767,7 +1797,7 @@ static PyObject *
 Ppmd8Encoder_flush(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
 {
     PyObject *ret;
-    PPMD_outBuffer out;
+    OutBuffer out;
     BlocksOutputBuffer buffer;
     BufferWriter writer;
 
@@ -1778,11 +1808,14 @@ Ppmd8Encoder_flush(Ppmd8Encoder *self, PyObject *args, PyObject *kwargs)
         goto error;
     }
 
-    writer.Write = Write;
+    writer.Write = (void (*)(void *, Byte)) Writer;
     writer.outBuffer = &out;
     self->cPpmd8->Stream.Out = (IByteOut *) &writer;
-    Ppmd8_EncodeSymbol(self->cPpmd8, 0x01);  // endmark sequence
-    Ppmd8_EncodeSymbol(self->cPpmd8, 0x00);
+
+    if (self->endmark) {
+        Ppmd8_EncodeSymbol(self->cPpmd8, 0x01);  // endmark sequence
+        Ppmd8_EncodeSymbol(self->cPpmd8, 0x00);
+    }
     Ppmd8_EncodeSymbol(self->cPpmd8, -1);  // endmark for encoder
     Ppmd8_RangeEnc_FlushData(self->cPpmd8);
     ret = OutputBuffer_Finish(&buffer, &out);
@@ -1894,6 +1927,12 @@ PyInit__ppmd(void) {
     if (!module) {
         goto error;
     }
+    PyModule_AddIntConstant(module, "PPMD8_RESTORE_METHOD_RESTART", 0);
+    PyModule_AddIntConstant(module, "PPMD8_RESTORE_METHOD_CUT_OFF", 1);
+    // #ifdef PPMD8_FREEZE_SUPPORT
+    // PyModule_AddIntConstant(module, "PPMD8_RESTORE_METHOD_FREEZE", 2);
+    // #endif
+
     if (add_type_to_module(module,
                            "Ppmd7Encoder",
                            &Ppmd7Encoder_type_spec,
